from http import HTTPStatus

from .configuration import ConfigError
from .webtypes import Response, Request
from .router import Methods


class CORSHandler:
    __slots__ = ('allowed_origins', 'allowed_headers', 'allowed_methods')
    """
    Abstracts CORS things
    """
    def __init__(self, *, allowed_origins: set, allowed_headers: str,
                 allowed_methods: str):
        if len(allowed_origins) == 1:
            self.allowed_origins = allowed_origins.pop()
        else:
            self.allowed_origins = allowed_origins
        self.allowed_headers = allowed_headers
        self.allowed_methods = allowed_methods

    def add_cors_headers(self, request: Request, response: Response):
        if isinstance(self.allowed_origins, str):
            # only one origin
            result = self.allowed_origins

        elif request.headers.get('origin') in self.allowed_origins:
            result = request.headers.get('origin')

        else:
            result = 'none'

        response.headers['Access-Control-Allow-Origin'] = result
        response.headers['Access-Control-Allow-Credentials'] = 'true'

        if request.method == Methods.OPTIONS:
            if self.allowed_headers:
                response.headers['Access-Control-Allow-Headers'] = \
                    self.allowed_headers
            if self.allowed_methods:
                response.headers['Access-Control-Allow-Methods'] = \
                    self.allowed_methods

    async def options_handler(self, request):
        return Response(status=HTTPStatus.NO_CONTENT)

    @staticmethod
    def from_config(config):
        try:
            config['cors']['handle']
        except ConfigError:
            return None

        origins = config['cors']['allowed_origins']
        if not origins:  # if empty
            raise ValueError('Must have at least one cors.allowed_origins'
                             ' in your configuration.')

        origins = {origin.strip() for origin in origins.split(',')}

        try:
            headers = config['cors']['allowed_headers']
        except ConfigError:
            headers = ''

        try:
            methods = config['cors']['allowed_methods']
        except ConfigError:
            methods = ''

        return CORSHandler(allowed_origins=origins, allowed_headers=headers,
                           allowed_methods=methods)



import os
import yaml

CONFIG_LOCATION = os.getenv('WASPY_CONFIG_LOCATION')


class ConfigError(KeyError):
    """ Raised when a requested configuration can not be found """
    def __init__(self, config_name, env_var):
        super().__init__("""\
No configuration found for "{}". \
Add the environment variable {} or add \
the key to your config.yaml file.\
""".format(config_name, env_var))

NO_CONFIG_LOCATION_ERROR_MESSAGE = """
No config file specified. \
You can use `app.config.from_file(file_location)` or set a location using \
the environment variable "WASPY_CONFIG_LOCATION".\
"""

CONFIG_NOT_YET_LOADED_ERROR_MESSAGE = """
Configuration file not yet loaded. \
You should use `[app.]config.from_file(location)` or `[app.]config.load` \
before trying to access a configuration value.\
"""


class Config:
    """ load configuration from envvar or config file
        You can get settings from environment variables or a config.yaml file
        Environment variables have a higher precedence over the config file

        Use `config['section']['subsection']['key']` to get the value
        SECTION_SUBSECTION_KEY from environment variables or
        section.subsection.key from a yaml file
        (usually written:
            [section.subsection]
            key = value
        in a yaml file)
    """
    def __init__(self, _basename=None, _defaults=None):
        self.default_options = _defaults
        self.basename = _basename

    def from_file(self, location):
        self._load_config(filepath=location)
        return self

    def load(self):
        if self.default_options is None:
            self._load_config()

    def _load_config(self, filepath=None):
        if filepath is None:
            if CONFIG_LOCATION is None:
                raise ValueError(NO_CONFIG_LOCATION_ERROR_MESSAGE)
            filepath = os.path.abspath(CONFIG_LOCATION)
        with open(filepath, 'r') as f:
            config = yaml.safe_load(f)
            self.default_options = config

    def __getitem__(self, item):
        if self.default_options is None:
            raise ValueError(CONFIG_NOT_YET_LOADED_ERROR_MESSAGE)
        default = self.default_options.get(item)

        if isinstance(default, dict):
            return Config(_basename=self._create_basename(item),
                          _defaults=default)

        env = self._get_env_var(item)
        if env is None:
            if default is None:
                raise ConfigError(self._create_basename(item),
                                  self._create_env_var_string(item))
            return default
        return env

    def _create_basename(self, item):
        if self.basename:
            return self.basename + '.' + item
        else:
            return item

    def _create_env_var_string(self, item):
        if not self.basename:
            return item.upper()
        else:
            return '_'.join(self._create_basename(item).split('.')).upper()

    def _get_env_var(self, item):
        envvar_string = self._create_env_var_string(item)
        env = os.getenv(envvar_string)
        if env is not None:
            if isinstance(env, str):
                # env vars are always passed in as strings in docker world
                # here we will try to convert them to basic types if we can
                if env.lower() == 'true':
                    return True
                if env.lower() == 'false':
                    return False
                try:
                    env = int(env)
                except ValueError:
                    pass
        return env




from .httptransport import HTTPTransport as HTTPToolsTransport
from .httptransport import HTTPTransport, HTTPClientTransport
from .rabbitmqtransport import RabbitMQTransport, RabbitMQClientTransport
from .transportabc import TransportABC
from .testtransport import TestTransport




import asyncio
import json

from .. import webtypes
from .transportabc import TransportABC, ClientTransportABC


class TestClientTransport(ClientTransportABC):
    async def make_request(self, service: str, method: str, path: str,
                     body: bytes = None, query: str = None,
                     headers: dict = None, correlation_id: str = None,
                     content_type: str = None, **kwargs) -> webtypes.Response:
        pass


class TestTransport(TransportABC):
    def __init__(self, *args, **kwargs):
        self.app = None
        self.loop = None
        self.handler = None

    def listen(self, *, loop, config):
        self.loop = loop

    def run_app(self, app):
        self.app = app
        app.transport = (self,)
        app.shutdown = lambda: ''
        app.run()
        # now patch app to have send_request methods as well
        def send_request_for_app(request):
            return self.send_request(request)
        async def send_async_request_for_app(request):
            return await self.send_async_request(request)

        app.send_request = send_request_for_app
        app.send_async_request = send_async_request_for_app

    def send_request(self, request):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.send_async_request(request))

    async def send_async_request(self, request):
        if request.body and isinstance(request.body, dict):
            request.body = json.dumps(request.body)

        if isinstance(request.body, str):
            request.body = request.body.encode()

        response = await self.handler(request)
        response.body = response.data
        return response

    async def start(self, request_handler: callable):
        self.handler = request_handler

    def shutdown(self):
        pass

    def get_client(self):
        return TestClientTransport()



import asyncio
import traceback
import logging
import urllib.parse
from http import HTTPStatus

from httptools import HttpRequestParser, HttpResponseParser, HttpParserError, \
        parse_url

from ..webtypes import Request, Response
from .transportabc import TransportABC, ClientTransportABC

logger = logging.getLogger('waspy')


class ClosedError(Exception):
    """ Error for closed connections """


class _HTTPClientConnection:
    slots = ('reader', 'writer', 'http_parser', '_done', '_data')

    def __init__(self):
        self.reader = None
        self.writer = None
        self.http_parser = HttpResponseParser(self)
        self.response = None
        self._data = b''
        self._done = False

    async def connect(self, service, port, use_ssl):
        self.reader, self.writer = await \
            asyncio.open_connection(service, port, ssl=use_ssl)

    def send(self, method, path, headers, body):
        self.writer.write(f'{method.upper()} {path} HTTP/1.0\r\n'
                          .encode('latin-1'))
        for header, value in headers:
            self.writer.write(f'{header}: {value}\r\n'.encode('latin-1'))
        self.writer.write(b'\r\n')
        if body:
            self.writer.write(body)

    async def get_response(self):
        while True:
            data = await self.reader.read(1064)
            self.http_parser.feed_data(data)
            if self._done:
                return self.response

    def close(self):
        self.writer.close()

    """ http parsing methods below """

    def on_message_begin(self):
        self.response = Response()

    def on_header(self, name, value):
        name = name.decode('latin-1')
        value = value.decode()
        if name == 'X-Correlation-ID':
            self.response.correlation_id = value
        elif name.lower() == 'content-type':
            self.response.content_type = value
        else:
            self.response.headers[name] = value

    def on_headers_complete(self):
        self.response.status = HTTPStatus(self.http_parser.get_status_code())

    def on_body(self, body):
        self._data += body

    def on_message_complete(self):
        self.response.body = self._data
        self._data = b''
        self._done = True


class HTTPClientTransport(ClientTransportABC):
    """Client implementation of the HTTP transport protocol"""
    def _get_connection_for_service(self, service):
        pass

    async def make_request(self, service, method, path, body=None, query=None,
                           headers=None, correlation_id=None,
                           content_type=None, port=80, **kwargs):
        # form request object
        if not path.startswith('/'):
            path = '/' + path
        if headers is None:
            headers = {}
        use_ssl = service.startswith('https://') or port == 443
        if service.startswith('http'):
            service = service.replace('http://', '').replace('https://', '')
        if use_ssl and port == 80:
            port = 443
        if 'Host' not in headers and 'host' not in headers:
            headers['Host'] = service
            if port not in (80, 443):
                headers['Host'] += ':{}'.format(port)
        headers['Connection'] = 'close'
        headers.pop('connection', None)
        if correlation_id:
            headers['X-Correlation-Id'] = correlation_id
        if query:
            path += '?' + query
        if content_type:
            headers['Content-Type'] = content_type
        headers.pop('content-type', None)
        headers.pop('content-length', None)
        headers.pop('Content-Length', None)
        if body:
            headers['Content-Length'] = str(len(body))
        headers['User-Agent'] = headers.pop('user-agent', 'waspy-http-client')

        # now make a connection and send it
        connection = _HTTPClientConnection()
        await connection.connect(service, port, use_ssl)
        connection.send(method, path, headers.items(), body)
        try:
            result = await connection.get_response()
        finally:
            connection.close()
        return result


class HTTPTransport(TransportABC):
    """ Server implementation of the HTTP transport protocol"""

    def get_client(self):
        return HTTPClientTransport()

    def __init__(self, port=8080, prefix=None, shutdown_grace_period=5,
                 shutdown_wait_period=1):
        """
         HTTP Transport for listening on http
         :param port: The port to lisen on (0.0.0.0 will always be used)
         :param prefix: the path prefix to remove from all url's
         :param shutdown_grace_period: Time to wait for server to shutdown
         before connections get forceably closed. The only way for connections
         to not be forcibly closed is to have some connection draining in front
         of the service for deploys. Most docker schedulers will do this for you.
         :param shutdown_wait_period: Time to wait after recieving the sigterm
         before starting shutdown 
         """
        self.port = port
        if prefix is None:
            prefix = ''
        self.prefix = prefix
        self._handler = None
        self._server = None
        self._loop = None
        self._done_future = asyncio.Future()
        self._connections = set()
        self.shutdown_grace_period = shutdown_grace_period
        self.shutdown_wait_period = shutdown_wait_period
        self.shutting_down = False

    def listen(self, *, loop: asyncio.AbstractEventLoop, config):
        self._loop = loop
        self._config = config
        if self._config['debug']:
            self.shutdown_grace_period = 0
            self.shutdown_wait_period = 0
            self._debug = True

    async def start(self, request_handler):
        self._handler = request_handler
        self._server = await self._loop.create_server(
            lambda: _HTTPServerProtocol(parent=self, loop=self._loop),
            host='0.0.0.0',
            port=self.port,
            reuse_address=True,
            reuse_port=True)
        print(f'-- Listening for HTTP on port {self.port} --')
        try:
            await self._done_future
        except asyncio.CancelledError:
            pass

        logger.warning("Shutting down HTTP transport")
        await asyncio.sleep(self.shutdown_wait_period)
        # wait for connections to stop
        times_no_connections = 0
        for _ in range(self.shutdown_grace_period):
            if not self._connections:
                times_no_connections += 1
            else:
                times_no_connections = 0
                for con in self._connections:
                    con.attempt_close()

            if times_no_connections > 3:
                # three seconds with no connections
                break
            await asyncio.sleep(1)

        # Shut the server down
        self._server.close()
        await self._server.wait_closed()

    async def handle_incoming_request(self, request):
        logger.debug('received incoming request via http: %s', request)
        response = await self._handler(request)
        return response

    def shutdown(self):
        self.shutting_down = True
        self._done_future.cancel()


class _HTTPServerProtocol(asyncio.Protocol):
    """ HTTP Protocol handler.
        Should only be used by HTTPServerTransport
    """
    __slots__ = ('_parent', '_transport', '_task', 'data', 'http_parser',
                 'request')

    def __init__(self, *, parent, loop):
        self._parent = parent
        self._transport = None
        self.data = None
        self.http_parser = HttpRequestParser(self)
        self.request = None
        self._loop = loop
        self._task: asyncio.Task = None

    """ The next 3 methods are for asyncio.Protocol handling """
    def connection_made(self, transport):
        self._transport = transport
        self._parent._connections.add(self)

    def connection_lost(self, exc):
        self._parent._connections.discard(self)
        if self._task:
            self._task.cancel()
        self._transport = None

    def data_received(self, data):
        try:
            self.http_parser.feed_data(data)
        except HttpParserError as e:
            traceback.print_exc()
            logger.error('Bad http: %s', self.request)
            if self._transport:
                self.send_response(Response(status=400,
                                            body={'reason': 'Invalid HTTP',
                                                  'details': str(e)}))

    """ 
    The following methods are for HTTP parsing (from httptools)
    """
    def on_message_begin(self):
        self.request = Request()
        self.data = b''

    def on_header(self, name, value):
        key = name.decode('latin-1').lower()
        if not value:
            value = b''

        val = value.decode()
        self.request.headers[key] = val
        if key == 'x-correlation-id':
            self.request.correlation_id = val
        if key == 'content-type':
            self.request.content_type = val

    def on_headers_complete(self):
        self.request.method = self.http_parser.get_method().decode('latin-1')

    def on_body(self, body: bytes):
        self.data += body

    def on_message_complete(self):
        self.request.body = self.data
        task = self._loop.create_task(
            self._parent.handle_incoming_request(self.request)
        )
        task.add_done_callback(self.handle_response)
        self._task = task

    def on_url(self, url):
        url = url.replace(b'//', b'/')
        url = parse_url(url)
        if url.query:
            # query = urllib.parse.unquote(url.query.decode('latin-1'))
            self.request.query_string = url.query.decode('latin-1')
        path = urllib.parse.unquote(url.path.decode('latin-1'))
        if path.startswith(self._parent.prefix):
            path = path[len(self._parent.prefix):]
        self.request.path = path

    """
    End parsing methods
    """

    def handle_response(self, future):
        try:
            self.send_response(future.result())
        except Exception:
            traceback.print_exc()
            self.send_response(
                Response(status=500,
                         body={'reason': 'Something really bad happened'}))

    def send_response(self, response):
        if response is None:
            # connection closed, no response
            return

        headers = 'HTTP/1.1 {status_code} {status_message}\r\n'.format(
            status_code=response.status.value,
            status_message=response.status.phrase,
        )
        if self._parent.shutting_down:
            headers += 'Connection: close\r\n'
        else:
            headers += 'Connection: keep-alive\r\n'
            headers += 'Keep-Alive: timeout=5, max=50\r\n'

        if response.data:
            headers += 'Content-Type: {}\r\n'.format(response.content_type)
            headers += 'Content-Length: {}\r\n'.format(len(response.data))
            if ('transfer-encoding' in response.headers or
                        'Transfer-Encoding' in response.headers):
                print('Httptoolstransport currently doesnt support '
                      'chunked mode, attempting without.')
                response.headers.pop('transfer-encoding', None)
                response.headers.pop('Transfer-Encoding', None)
        else:
            headers += 'Content-Length: {}\r\n'.format(0)
        for header, value in response.headers.items():
            if header in ('Content-Length', 'content-lenth'):
                continue
            headers += '{header}: {value}\r\n'.format(header=header,
                                                      value=value)

        result = headers.encode('latin-1') + b'\r\n'
        if response.data:
            result += response.data

        try:
            self._transport.write(result)
        except AttributeError:
            # "NoneType has no attribute 'write'" because transport is closed
            logger.debug('Connection closed prematurely, most likely by client')
        self.request = 0
        self.data = 0

    def attempt_close(self):
        if self.request == 0:
            self._transport.close()



import json
from collections import defaultdict
from urllib import parse
from aenum import extend_enum
from .router import Methods
from http import HTTPStatus, cookies

extend_enum(HTTPStatus, 'INVALID_REQUEST', (430, 'Invalid Request',
                                            'Request was syntactically sound, '
                                            'but failed validation rules'))


class QueryParams:
    """
    A dictionary that stores multiple values per key.

    this has all the normal dictionary methods, and works as normal but does
    not override a key when `add` is used, and also has `getall`
    """
    __slots__ = ['mappings']

    @classmethod
    def from_string(cls, string):
        query_params = QueryParams()
        qs = parse.parse_qsl(string)
        for k, v in qs:
            query_params.add(k, v)

        return query_params

    def __init__(self):
        self.mappings = defaultdict(list)

    def get(self, name, default=None):
        return self.mappings.get(name, [default])[0]

    def getall(self, name, default=None):
        return self.mappings.get(name, default)

    def __getitem__(self, key):
        try:
            return self.mappings.__getitem__(key)[0]
        except IndexError:
            raise KeyError('Invalid Key: {key}'.format(key))

    def __setitem__(self, key, value):
        raise TypeError('MultiDict does not support item assignment. '
                        'Use .add(k, v) instead.')

    def add(self, name, value):
        self.mappings[name].append(value)

    def __str__(self):
        return parse.urlencode(self.mappings, doseq=True)


class Request:
    def __init__(self, headers: dict = None,
                 path: str = None, correlation_id: str = None,
                 method: str = None, query_string: str = None,
                 body: bytes=None, content_type='application/json'):

        if not headers:
            headers = {}
        if not method:
            method = 'GET'
        self._method = None
        self.headers = headers
        self.path = path
        self.correlation_id = correlation_id
        self.method = method  # this is a property setter
        self.query_string = query_string
        self._query_params = None
        self.body = body
        self.path_params = {}
        self._handler = None
        self.app = None
        self.content_type = content_type
        self._json = None
        self._cookies = None

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, value):
        if isinstance(value, str):
            value = Methods(value.upper())
        self._method = value

    @property
    def path_qs(self):
        # the path + the query string
        query = '?' + self.query_string if self.query_string else ''
        return self.path + query

    @property
    def cookies(self) -> dict:
        if self._cookies is None:
            self._cookies = {}
            raw = self.headers.get('cookie', None)
            if raw:
                cookie_manager = cookies.SimpleCookie(raw)
                self._cookies = {i: cookie_manager[i].value for i in cookie_manager}
        return self._cookies

    @property
    def query(self) -> QueryParams:
        # parse query string into a dictionary
        if not self._query_params:
            self._query_params = QueryParams.from_string(self.query_string)
        return self._query_params

    def json(self) -> dict:
        if not self._json:
            # convert body into a json dict
            try:
                self._json = json.loads(self.body.decode())
            except json.JSONDecodeError as ex:
                raise JSONDecodeError from ex
        return self._json

    def __str__(self):
        query = '?' + self.query_string if self.query_string else ''
        return('<Request({method} {path}{query})@{id}>'
               .format(method=self.method, path=self.path,
                       query=query, id=id(self)))

    def __repr__(self):
        return (f'Request(headers={repr(self.headers)}, path={repr(self.path)}, '
                f'correlation_id={self.correlation_id}, method={repr(self.method)}, '
                f'query_string={self.query_string}, body={self.body}, '
                f'content_type={self.content_type})')


class Response:
    def __init__(self, headers=None, correlation_id=None,
                 body=None, status=HTTPStatus.OK,
                 content_type='application/json', meta: dict=None):
        """
        Response object
        :param headers:
        :param correlation_id:
        :param body: message body
        :param status: status code
        :param content_type:
        :param meta: Extra context information.
            Not to be returned through transport
        """
        if not headers:
            headers = dict()
        if isinstance(status, int):  # convert to enum
            status = HTTPStatus(status)
        if meta is None:
            meta = {}
        self.headers = headers
        self.correlation_id = correlation_id
        self.body = body
        self.status = status
        self.content_type = content_type
        self.meta = meta
        self._data = None
        self._json = None

    def __str__(self):
        return('<Response({status})@{id}>'
               .format(status=self.status, id=id(self)))

    @property
    def data(self):
        if self._data is None and self.body:
            if (self.content_type == 'application/json' and
                    isinstance(self.body, dict)):
                self._data = json.dumps(self.body)
            else:
                self._data = self.body
            if isinstance(self._data, str):
                self._data = self._data.encode()
        return self._data

    def json(self) -> dict:
        """
        Used for client response
        """
        if self.body is None:
            self._json = {}
        elif not self._json:
            # convert body into a json dict
            try:
                self._json = json.loads(self.body.decode())
            except json.JSONDecodeError as ex:
                raise JSONDecodeError from ex
        return self._json


class ResponseError(Exception):
    def __init__(self, message=None, status=None, *, body=None, headers=None,
                 correlation_id=None, reason=None, log=False):
        super().__init__(message)
        self.message = message
        if hasattr(self, 'status') and status is None:
            status = self.status
        if hasattr(self, 'body') and body is None:
            body = self.body
        if hasattr(self, 'reason') and reason is None:
            reason = self.reason
        if hasattr(self, 'log') and log == False:
            log = self.log
        if reason and not body:
            body = {'reason': reason}
        self.response = Response(status=status, body=body, headers=headers,
                                 correlation_id=correlation_id)
        self.log = log


class JSONDecodeError(ResponseError):
    status = 400
    reason = 'Invalid Json'


class NotRoutableError(ResponseError):
    status = 404
    reason = 'No route found'



import asyncio
import logging
import signal
import sys
from functools import wraps
from typing import List, Union, Iterable
from http import HTTPStatus
from concurrent.futures import CancelledError

from ._cors import CORSHandler
from .client import Client
from .webtypes import Request, Response, ResponseError
from .router import Router
from .transports.transportabc import TransportABC
from .transports.rabbitmqtransport import NackMePleaseError
from .configuration import Config, ConfigError
from . import errorlogging


logging.basicConfig(format='%(asctime)s %(levelname)s [%(module)s.%(funcName)s] %(message)s')
logger = logging.getLogger('waspy')


async def response_wrapper_factory(app, handler):
    @wraps(handler)
    async def wrap_response_middleware(request):
        response = await handler(request)
        if not isinstance(response, Response):
            if isinstance(response, tuple):
                body = response[0]
                status = response[1]
                response = Response(status=status, body=body)
            elif isinstance(response, dict) or isinstance(response, str):
                response = Response(body=response)
            elif response is None:
                response = Response(status=HTTPStatus.NO_CONTENT)
            else:
                raise ValueError('Request handler returned an invalid type.'
                                 ' Return types should be one of '
                                 '[Response, dict, str, None, (dict, int)]')
        return response
    return wrap_response_middleware


class Application:
    def __init__(self,
                 transport: Union[TransportABC,
                                  Iterable[TransportABC]]=None,
                 *,
                 middlewares: List[callable]=None,
                 default_headers: dict=None,
                 debug: bool=False,
                 router: Router=None,
                 config: Config=None,
                 loop=None):
        if transport is None:
            from waspy.transports.httptransport import HTTPTransport
            transport = HTTPTransport()
        if isinstance(transport, list):   
            transport = tuple(transport)
        if not isinstance(transport, tuple):
            transport = (transport,)
        if middlewares is None:
            middlewares = ()
        middlewares = tuple([m for m in middlewares])
        middlewares += (response_wrapper_factory,)
        if router is None:
            router = Router()
        if default_headers is None:
            default_headers = {'Server': 'waspy'}
        if not config:
            config = Config()
        self.transport = transport
        self.middlewares = middlewares
        self.default_headers = default_headers
        self.debug = debug
        self.router = router
        self.on_start = []
        self.on_stop = []
        self._client = None
        self.config = config
        self.raven = None
        self.logger = None
        self._cors_handler = None
        self.loop = loop

    @property
    def client(self) -> Client:
        if not self._client:
            self._client = Client(transport=self.transport[0].get_client())
        return self._client

    def start_shutdown(self, signum=None, frame=None):
        # loop = asyncio.get_event_loop()
        for t in self.transport:
            t.shutdown()

    def run(self):
        if not self.loop:
            self.loop = asyncio.get_event_loop()
        loop = self.loop

        if self.config['debug']:
            logger.setLevel('DEBUG')
            self.loop.set_debug(True)

        # init logger
        self._create_logger()

        # add cors support if needed
        self._cors_handler = CORSHandler.from_config(self.config)
        if self._cors_handler:
            self.router.add_generic_options_handler(self._cors_handler.options_handler)

        # wrap handlers in middleware
        loop.run_until_complete(self._wrap_handlers())
        for t in self.transport:
            t.listen(loop=loop, config=self.config)

        # Call on-startup hooks
        loop.run_until_complete(self.run_on_start_hooks())

        # todo: fork/add processes?
        tasks = []
        for t in self.transport:
            tasks.append(t.start(self.handle_request))

        # register signals, so that stopping the service works correctly
        loop.add_signal_handler(signal.SIGTERM, self.start_shutdown)
        loop.add_signal_handler(signal.SIGINT, self.start_shutdown)

        # Run all transports - they shouldn't return until shutdown
        loop.run_until_complete(asyncio.gather(*tasks))

        self.shutdown()

    async def run_on_start_hooks(self):
        """
        Run all hooks in on_start. Allows for coroutines and synchronous functions.
        """
        logger.debug("Running on start hooks")
        await self._run_hooks(self.on_start)

    async def run_on_stop_hooks(self):
        """
        Run all hooks in on_stop. Allows for coroutines and synchronous functions.
        """
        logger.debug("Running on stop hooks")
        await self._run_hooks(self.on_stop)


    async def handle_request(self, request: Request) -> Response:
        """
        coroutine: This method is called by Transport
        implementation to handle the actual request.
        It returns a webtype.Response object.
        """
        # Get handler
        try:
            try:
                handler = self.router.get_handler_for_request(request)
                request.app = self
                response = await handler(request)

            except ResponseError as r:
                response = r.response
                if r.log:
                    exc_info = sys.exc_info()
                    self.logger.log_exception(request, exc_info, level='warning')
            # invoke serialization (json) to make sure it works
            _ = response.data

        except CancelledError:
            # This error can happen if a client closes the connection
            # The response shouldnt really ever be used
            return None

        except NackMePleaseError:
            """ See message where this error is defined """
            raise

        except Exception:
            exc_info = sys.exc_info()
            self.logger.log_exception(request, exc_info)
            response = Response(status=500,
                                body={'message': 'Server Error'})
        if not response.correlation_id:
            response.correlation_id = request.correlation_id

        if self._cors_handler is not None:
            self._cors_handler.add_cors_headers(request, response)

        # add default headers
        response.headers = {**self.default_headers, **response.headers}

        return response

    async def _wrap_handlers(self):
        handler_gen = self.router._get_and_wrap_routes()

        try:
            handler = next(handler_gen)
            while True:
                wrapped = handler
                for middleware in self.middlewares[::-1]:
                    wrapped = await middleware(self, wrapped)
                handler = handler_gen.send(wrapped)
        except StopIteration:
            pass

    def _create_logger(self):
        try:
            dsn = self.config['sentry']['dsn']
        except (ConfigError, ValueError):
            self.logger = errorlogging.ErrorLoggingBase()
        else:
            try:
                env = self.config['app_env']
            except (ConfigError):
                env = 'waspy'
            self.logger = errorlogging.SentryLogging(
                dsn=dsn,
                environment=env
            )

    async def _run_hooks(self, hooks):
        coros = []
        while len(hooks):
            task = hooks.pop()
            if asyncio.iscoroutinefunction(task):
                coros.append(task(self))
            else:
                task(self)
        await asyncio.gather(*coros)

    def shutdown(self):
        self.loop.run_until_complete(self.run_on_stop_hooks())
        self.loop.close()
